#########################################################################
##  ST_CH.PRC
##  
##  This file contains the R procedure ST_CH
##  
##      written by
##  
##  Bruce E. Hansen
##  Department of Economics
##  Social Science Building
##  University of Wisconsin
##  Madison, WI 53706-1393
##  behansen@wisc.edu
##  http://www.ssc.wisc.edu/~bhansen/
##  
##  Note: the procedure calls the procedures
##  pv_sup, pv_exp, and pv_ave
##  which are contained in separate ASCII files.
##  
##  *************************************************************
##  
##  ST_CH
##  Procedure to calculate test statistics and p-values for linear
##  regression models subject to structural change.
##  
##  Format:
##  
##  out <- st_ch(y,x,t1,t2,h)
##  kest$out 
##  sup$out
##  exp$out
##  ave$out
##
##  Inputs:
##  y = dependent variable (nx1 vector).
##  x = indenpendent variables (nxk matrix).
##  t1 = starting breakpoint percentage, number in (0,1)
##       or starting breakpoint index, number in [k,T-k].
##  t2 = ending breakpoint percentage, number in (0,1)
##       or ending breakpoint index, number in [k,T-k].
##  h  = Heteroskedasticity-Consistent Covariance Matrix dummy
##       1 to calculate heteroskedasticity-consistent covariances
##       0 to calculate conventional covariances.
##  
##  Outputs:
##  kest = Quandt estimate of breakdate
##  sup  = Andrews/Quandt SupLM test
##  exp  = Andrews/Ploberger ExpLM test (c=infinity)
##  ave  = Andrews/Ploberger AveLM test (c=0)
##  
##  kest is a (k+1)x2 vector.
##    Estimates of Breakpoint.
##    The first row reports the index of the element (i.e., the 22nd observation),
##      while the second reports the index's relative place in the sample
##      (i.e., if n=200, it equals 22/200=.11).
##    The first element is the estimate of breakpoint obtained by allowing all
##      parameter to shift.  The remaining k elements contain the estimates of
##      the breakpoint when only a single parameter is allowed to shift.
##  
##  Each of (sup, exp, ave) are (k+1)x2 matrices.
##    The first column contains test statistics, the second contains
##      asymptotic p-values calcuated from by the procedures pv_sup,
##      pv_exp, and pv_ave.
##    The first row reports the joint test for constancy of all regression
##      parameters.
##    The remaining nrow report the tests for partial constancy of each
##      regression parameter, separately.
##  
##  Note:
##  For Sup test, Andrews (1993) recommends t1 = .15 and t2=.85.
##  For Exp and Ave tests, Andrews-Ploberger (1994) recommend t1 = .02 and t2=.98.
##  
#########################################################################

source ("pv_sup.R")
source ("pv_exp.R")
source ("pv_ave.R")

st_ch <- function(y,x,t1,t2,h){

  xx <- t(x)%*%x  
  xxi <- solve(xx) 
  e <- y - x%*%xxi%*%(t(x)%*%y)
  k  <- ncol(x)
  n  <- nrow(e)
  sig <- (t(e)%*%e)/(n-k)
  if (h==1) xe <- x*(e%*%matrix(1,1,k)) else xe <- x*as.vector(sqrt(sig))
  xes <- x*(e%*%matrix(1,1,k))
  for (j in 1:k) xes[,j] <- as.matrix(cumsum(xes[,j]))
  mgm <- xxi%*%(t(xe)%*%xe)%*%xxi
  if (t1<1){
    n1 <- floor(n*t1)
    n2 <- floor(n*t2)
    tau1 <- t1
    tau2 <- t2
  }else{
    n1 <- t1
    n2 <- t2
    tau1 <- t1/n
    tau2 <- t2/n
  }

  if (n1<k){
    cat ("\n")
    cat ("\n")
    cat ("ERROR:  Starting Sample is smaller than Number of Parameters","\n")
    cat ("        You need to select a larger value for t1","\n")
    cat ("\n")
    cat ("\n")
  }
  if (n2>(n-k)){
    cat ("\n")
    cat ("\n")
    cat ("ERROR:  Ending Sample is smaller than Number of Paramters","\n")
    cat ("        You need to select a smaller value for t2","\n")
    cat ("\n")
    cat ("\n")
  }

  lm <- matrix(0,n,k+1) 
  m <- t(x[1:(n1-1),])%*%x[1:(n1-1),]
  g <- t(xe[1:(n1-1),])%*%xe[1:(n1-1),]

  for (i in n1:n2){
    m <- m + as.matrix(x[i,])%*%t(as.matrix(x[i,]))
    g <- g + as.matrix(xe[i,])%*%t(as.matrix(xe[i,]))
    mi <- g - m%*%xxi%*%g - g%*%xxi%*%m + m%*%mgm%*%m
    sn <- as.matrix(xes[i,])
    pw <- (sn^2)/as.matrix(diag(mi))
    jw <- t(sn)%*%solve(mi)%*%sn
    lm[i,] <- cbind(jw,t(pw))
  }

  supt <- as.matrix(apply(lm,2,max)) 
  kest <- as.matrix(max.col(t(lm)))
  kest <- cbind(kest,(kest/n))
  lmt <- lm[n1:n2,]
  expt <- as.matrix(log(colMeans(exp(lmt/2))))
  avet <- as.matrix(colMeans(lmt))

  pi0 <- 1/(1+sqrt(tau2*(1-tau1)/tau1/(1-tau2)))

  psup <- pv_sup(supt[1],k,pi0)
  pexp <- pv_exp(expt[1],k,pi0)
  pave <- pv_ave(avet[1],k,pi0)

  for (i in 1:k){
    psup <- rbind(psup,pv_sup(supt[i+1],1,pi0))
    pexp <- rbind(pexp,pv_exp(expt[i+1],1,pi0))
    pave <- rbind(pave,pv_ave(avet[i+1],1,pi0))
  }

list(kest=kest,sup=cbind(supt,psup),exp=cbind(expt,pexp),ave=cbind(avet,pave))
}


